#include <jni.h>
#include "JNISamples.h"
#include <iostream.h>

static void printString(JNIEnv *env, jstring jstr);

JNIEXPORT void JNICALL Java_JNISamples_syncNativeMethod (JNIEnv *env, jobject obj, jobject syncObj)
{
	cout << "Inside synchronized native method\n";
	env->MonitorEnter(syncObj);
	cout << "After synchronizing syncObj\n";
	env->MonitorExit(syncObj);
	cout << "After releasing syncObj\n";
	/*
	The functions Object.wait, Object.notify, and Object.notifyAll provide 
	another useful thread synchronization mechanism. While the JNI does not 
	directly support these functions, a native method can always follow the 
	JNI method call mechanism to invoke them.
	*/
}

JNIEXPORT void JNICALL Java_JNISamples_nativeCatchThrow (JNIEnv *env, jobject obj)
{
	jclass cls = env->GetObjectClass(obj);
	jmethodID mid = env->GetMethodID(cls, "javaCatchThrow", "()V");
	jthrowable exc;
	jclass newException;

	if(0 == mid)
	{
		cout << "Error obtaining method id\n";
		return;
	}

	env->CallVoidMethod(obj, mid);
	exc = env->ExceptionOccurred();

	if(exc)
	{
		/* 
		We don't do much with the exception, except that we print a debug message 
		using ExceptionDescribe, clear it, and throw a new exception. 
		*/
		/* 
		It is extremely important to check, handle, and clear the pending exception 
		before calling subsequent JNI functions. Calling arbitrary JNI functions with 
		a pending exception may lead to unexpected results. You can safely call only 
		a small number of JNI functions when there is a pending exception. These 
		functions are ExceptionOccurred, ExceptionDescribe, and ExceptionClear.
		*/

		env->ExceptionDescribe();
	    env->ExceptionClear();
	}

	newException = env->FindClass("java/lang/IllegalArgumentException");
	if (newException == 0) 
	{ 
		/* Unable to find the new exception class, give up. */
		cout << "Error finding the new exception class\n";
	    return;
	}

	env->ThrowNew(newException, "IllegalArgumentException thrown from C code");
}

JNIEXPORT void JNICALL Java_JNISamples_fieldAccess (JNIEnv *env, jobject obj)
{
	jfieldID fid;
	jstring jstr;
	const char *str;
	jint iStaticIntValue;
	jclass cls = env->GetObjectClass(obj);

	fid = env->GetStaticFieldID(cls, "iStaticIntFld", "I");
	if(0 == fid)
	{
		cout << "Error getting fid for iStaticIntFld\n";
		return;
	}
	iStaticIntValue = env->GetStaticIntField(cls, fid);
	cout << "iStaticIntValue=" << iStaticIntValue << "\n";
	env->SetStaticIntField(cls, fid, 200);

	fid = env->GetFieldID(cls, "sStringFld", "Ljava/lang/String;");
	if(0 == fid)
	{
		cout << "Error getting fid for sStringFld\n";
		return;
	}
	jstr = (jstring)env->GetObjectField(obj, fid);
	printString(env, jstr);

	jstr = env->NewStringUTF("NewString");
	env->SetObjectField(obj, fid, jstr);
}


JNIEXPORT void JNICALL Java_JNISamples_nativeMethod (JNIEnv *env, jobject obj, jint iDepth)
{
	jclass cls = env->GetObjectClass(obj);
	static jmethodID method = 0;

	/* 
	Method ID is valid only for as long as the class 
	from which it is derived is not unloaded.
	Obtaining a method ID is a relatively expensive operation.

	Use NewGlobalRef to create a global ref in native code to
	prevent the class from being garbage collected.
	You MUST use DeleteGlobalRef to delete the global reference.

	Use DeleteLocalRef to delete a local ref to release mem immediately.
	*/
	if(0 == method)
	{
		method = env->GetMethodID(cls, "callback", "(I)V");
	}

	if (method == 0)
	{
		cout << "Could not get method callback(I)V\n";
		return;
	}

	cout << "In C, depth = " << iDepth << "about to enter Java\n";
	env->CallVoidMethod(obj, method, iDepth);
	cout << "In C, depth = " << iDepth << ", back from Java\n";
}

JNIEXPORT jint JNICALL Java_JNISamples_stringArrAccess(JNIEnv *env, jobject obj, jobjectArray asArr)
{
	int iIndex;
	jsize len = env->GetArrayLength(asArr);

	for(iIndex=0; iIndex < len; iIndex++)
	{
		jstring jstr = (jstring)env->GetObjectArrayElement(asArr, iIndex);
		printString(env, jstr);
		/* explicitly releasing to assist garbage collection, though not required */
		env->DeleteLocalRef(jstr); 
	}

	return len;
}

JNIEXPORT jintArray JNICALL Java_JNISamples_intArrAccess (JNIEnv *env, jobject obj, jintArray aiArr)
{
	int i, sum = 0;
	jintArray aiRet;
	jint *retBody;

	jsize len = env->GetArrayLength(aiArr);
	jint *body = env->GetIntArrayElements(aiArr, 0);
	/* to copy a region of the array only, use Get/Set<type>ArrayRegion functions */

	for (i=0; i<len; i++) 
	{
		sum += body[i];
		body[i]++;
	}

	aiRet = env->NewIntArray(len);
	retBody = env->GetIntArrayElements(aiRet, 0);

	for (i=0; i<len; i++)
	{
		retBody[i] = body[i];
	}

	env->ReleaseIntArrayElements(aiArr, body, 0);
	env->ReleaseIntArrayElements(aiRet, retBody, 0);

	return aiRet;	
}


JNIEXPORT jstring JNICALL Java_JNISamples_stringAccess (JNIEnv *env, jobject obj, jstring prompt)
{
	char buf[128];
	const char * str = env->GetStringUTFChars(prompt, 0);

	cout << "Inside Native: " << str << "\n";
	sprintf(buf, "Hello [%s][%d][%d]", 
			str, 
			env->GetStringUTFLength(prompt), 
			env->GetStringLength(prompt));
	env->ReleaseStringUTFChars(prompt, str);

	return env->NewStringUTF(buf);
}

static void printString(JNIEnv *env, jstring jstr)
{
	const char * str = env->GetStringUTFChars(jstr, 0);
	cout << str << "\n";
	env->ReleaseStringUTFChars(jstr, str);
}

